@@ -6,11 +6,11 @@
 #define GEMMI_MTZ_HPP_
 
 #include <cassert>
-#include <cstdint>   // for int32_t
-#include <cstdio>    // for FILE, fprintf
-#include <cstring>   // for memcpy
-#include <cmath>     // for isnan
-#include <algorithm> // for sort, any_of
+#include <cstdint>       // for int32_t
+#include <cstdio>        // for FILE, fprintf
+#include <cstring>       // for memcpy
+#include <cmath>         // for isnan
+#include <algorithm>     // for sort, any_of
 #include <array>
 #include <initializer_list>
 #include <string>
@@ -92,10 +92,10 @@ struct Mtz {
     bool has_data() const { return parent->has_data(); }
     int size() const { return has_data() ? parent->nreflections : 0; }
     size_t stride() const { return parent->columns.size(); }
-    float& operator[](int n) { return parent->data[idx + n * stride()]; }
-    float operator[](int n) const { return parent->data[idx + n * stride()]; }
-    float& at(int n) { return parent->data.at(idx + n * stride()); }
-    float at(int n) const { return parent->data.at(idx + n * stride()); }
+    float& operator[](std::size_t n) { return parent->data[idx + n * stride()]; }
+    float operator[](std::size_t n) const { return parent->data[idx + n * stride()]; }
+    float& at(std::size_t n) { return parent->data.at(idx + n * stride()); }
+    float at(std::size_t n) const { return parent->data.at(idx + n * stride()); }
     bool is_integer() const {
       return type == 'H' || type == 'B' || type == 'Y' || type == 'I';
     }
@@ -145,6 +145,8 @@ struct Mtz {
 
     int dataset_id() const { return ints[20]; }
     void set_dataset_id(int id) { ints[20] = id; }
+    float wavelength() const { return floats[86]; }
+    void set_wavelength(float lambda) { floats[86] = lambda; }
     float phi_start() const { return floats[36]; }
     float phi_end() const { return floats[37]; }
     Mat33 matrix_U() const {
@@ -156,7 +158,8 @@ struct Mtz {
 
   std::string source_path;  // input file path, if known
   bool same_byte_order = true;
-  std::int32_t header_offset = 0;
+  bool indices_switched_to_original = false;
+  std::int64_t header_offset = 0;
   std::string version_stamp;
   std::string title;
   int nreflections = 0;
@@ -180,11 +183,8 @@ struct Mtz {
   FILE* warnings = nullptr;
 
   explicit Mtz(bool with_base=false) {
-    if (with_base) {
-      datasets.push_back({0, "HKL_base", "HKL_base", "HKL_base", cell, 0.});
-      for (int i = 0; i != 3; ++i)
-        add_column(std::string(1, "HKL"[i]), 'H', 0, i);
-    }
+    if (with_base)
+      add_base();
   }
   Mtz(Mtz&& o) noexcept { *this = std::move(o); }
   Mtz& operator=(Mtz&& o) noexcept {
@@ -193,7 +193,7 @@ struct Mtz {
     version_stamp = std::move(o.version_stamp);
     title = std::move(o.title);
     nreflections = o.nreflections;
-    sort_order = std::move(o.sort_order);
+    sort_order = o.sort_order;
     min_1_d2 = o.min_1_d2;
     max_1_d2 = o.max_1_d2;
     valm = o.valm;
@@ -217,6 +217,12 @@ struct Mtz {
   Mtz(Mtz const&) = delete;
   Mtz& operator=(Mtz const&) = delete;
 
+  void add_base() {
+    datasets.push_back({0, "HKL_base", "HKL_base", "HKL_base", cell, 0.});
+    for (int i = 0; i != 3; ++i)
+      add_column(std::string(1, "HKL"[i]), 'H', 0, i, false);
+  }
+
   // Functions to use after MTZ headers (and data) is read.
 
   double resolution_high() const { return std::sqrt(1.0 / max_1_d2); }
@@ -241,16 +247,24 @@ struct Mtz {
   }
 
   UnitCell get_average_cell_from_batch_headers(double* rmsd) const {
+    if (rmsd)
+      for (int i = 0; i < 6; ++i)
+        rmsd[i] = 0.;
     double avg[6] = {0., 0., 0., 0., 0., 0.};
     for (const Batch& batch : batches)
-      for (int i = 0; i < 6; ++i)
+      for (int i = 0; i < 6; ++i) {
+        // if batch headers are not set correctly, return global cell
+        if (batch.floats[i] <= 0)
+          return cell;
         avg[i] += batch.floats[i];
+      }
+    if (avg[0] <= 0 || avg[1] <= 0 || avg[2] <= 0 ||
+        avg[3] <= 0 || avg[4] <= 0 || avg[5] <= 0)
+      return UnitCell();
     size_t n = batches.size();
     for (int i = 0; i < 6; ++i)
       avg[i] /= n;
     if (rmsd) {
-      for (int i = 0; i < 6; ++i)
-        rmsd[i] = 0.;
       for (const Batch& batch : batches)
         for (int i = 0; i < 6; ++i)
           rmsd[i] += sq(avg[i] - batch.floats[i]);
@@ -324,6 +338,37 @@ struct Mtz {
     return cols;
   }
 
+  std::vector<int> positions_of_columns_with_type(char col_type) const {
+    std::vector<int> cols;
+    for (int i = 0; i < (int) columns.size(); ++i)
+      if (columns[i].type == col_type)
+        cols.push_back(i);
+    return cols;
+  }
+
+  // F(+)/(-) pairs should have type G (and L for sigma),
+  // I(+)/(-) -- K (M for sigma), but E(+)/(-) has no special column type,
+  // so here we use column labels not types.
+  std::vector<std::pair<int,int>> positions_of_plus_minus_columns() const {
+    std::vector<std::pair<int,int>> r;
+    for (int i = 0; i < (int) columns.size(); ++i) {
+      const Column& col = columns[i];
+      size_t sign_pos = col.label.find("(+)");
+      if (sign_pos != std::string::npos) {
+        std::string minus_label = columns[i].label;
+        minus_label[sign_pos+1] = '-';
+        for (int j = 0; j < (int) columns.size(); ++j)
+          if (columns[j].label == minus_label &&
+              columns[j].type == col.type &&
+              columns[j].dataset_id == col.dataset_id) {
+            r.emplace_back(i, j);
+            break;
+          }
+      }
+    }
+    return r;
+  }
+
   const Column* column_with_one_of_labels(std::initializer_list<const char*> labels) const {
     for (const char* label : labels) {
       if (const Column* col = column_with_label(label))
@@ -344,8 +389,9 @@ struct Mtz {
   }
 
   Column* rfree_column() {
+    // cf. MtzToCif::default_spec in mtz2cif.hpp
     return column_with_type_and_one_of_labels('I',
-        {"FREE", "RFREE", "FREER", "FreeR_flag", "R-free-flags"});
+        {"FREE", "RFREE", "FREER", "FreeR_flag", "R-free-flags", "FreeRflag"});
   }
   const Column* rfree_column() const {
     return const_cast<Mtz*>(this)->rfree_column();
@@ -417,14 +463,14 @@ struct Mtz {
 
   void toggle_endiannes() {
     same_byte_order = !same_byte_order;
-    swap_four_bytes(&header_offset);
+    swap_eight_bytes(&header_offset);
   }
 
   template<typename Stream>
   void read_first_bytes(Stream& stream) {
-    char buf[12] = {0};
+    char buf[20] = {0};
 
-    if (!stream.read(buf, 12))
+    if (!stream.read(buf, 20))
       fail("Could not read the MTZ file (is it empty?)");
     if (buf[0] != 'M' || buf[1] != 'T' || buf[2] != 'Z' || buf[3] != ' ')
       fail("Not an MTZ file - it does not start with 'MTZ '");
@@ -439,9 +485,19 @@ struct Mtz {
     if ((buf[9] & 0xf0) == (is_little_endian() ? 0x10 : 0x40))
       toggle_endiannes();
 
-    std::memcpy(&header_offset, buf + 4, 4);
+    std::int32_t tmp_header_offset;
+    std::memcpy(&tmp_header_offset, buf + 4, 4);
     if (!same_byte_order)
-      swap_four_bytes(&header_offset);
+      swap_four_bytes(&tmp_header_offset);
+
+    if (tmp_header_offset == -1) {
+      std::memcpy(&header_offset, buf + 12, 8);
+      if (!same_byte_order) {
+        swap_eight_bytes(&header_offset);
+      }
+    } else {
+      header_offset = (int64_t) tmp_header_offset;
+    }
   }
 
   static const char* skip_word(const char* line) {
@@ -714,32 +770,37 @@ struct Mtz {
     if (input.is_stdin()) {
       read_stream(FileStream{stdin}, with_data);
     } else if (CharArray mem = input.uncompress_into_buffer()) {
-      read_stream(MemoryStream(mem.data(), mem.size()), with_data);
+      read_stream(mem.stream(), with_data);
     } else {
       fileptr_t f = file_open(input.path().c_str(), "rb");
       read_stream(FileStream{f.get()}, true);
     }
   }
 
-  std::vector<int> sorted_row_indices() const {
+  std::vector<int> sorted_row_indices(int use_first=3) const {
     if (!has_data())
       fail("No data.");
+    if (use_first <= 0 || use_first >= (int) columns.size())
+      fail("Wrong use_first arg in Mtz::sort.");
     std::vector<int> indices(nreflections);
     for (int i = 0; i != nreflections; ++i)
       indices[i] = i;
     std::stable_sort(indices.begin(), indices.end(), [&](int i, int j) {
       int a = i * (int) columns.size();
       int b = j * (int) columns.size();
-      return data[a] < data[b] || (data[a] == data[b] && (
-               data[a+1] < data[b+1] || (data[a+1] == data[b+1] && (
-                 data[a+2] < data[b+2]))));
+      for (int n = 0; n < use_first; ++n)
+        if (data[a+n] != data[b+n])
+          return data[a+n] < data[b+n];
+      return false;
     });
     return indices;
   }
 
-  bool sort() {
-    std::vector<int> indices = sorted_row_indices();
-    sort_order = {{1, 2, 3, 0, 0}};
+  bool sort(int use_first=3) {
+    std::vector<int> indices = sorted_row_indices(use_first);
+    sort_order = {{0, 0, 0, 0, 0}};
+    for (int i = 0; i < use_first; ++i)
+      sort_order[i] = i + 1;
     if (std::is_sorted(indices.begin(), indices.end()))
       return false;
     std::vector<float> new_data(data.size());
@@ -758,8 +819,71 @@ struct Mtz {
       data[offset + i] = static_cast<float>(hkl[i]);
   }
 
+  // (for merged MTZ only) change HKL to ASU equivalent, adjust phases
+  void ensure_asu(bool tnt_asu=false) {
+    if (!is_merged())
+      fail("Mtz::ensure_asu() is for merged MTZ only");
+    if (!spacegroup)
+      return;
+    GroupOps gops = spacegroup->operations();
+    ReciprocalAsu asu(spacegroup, tnt_asu);
+    std::vector<int> phase_columns = positions_of_columns_with_type('P');
+    std::vector<int> abcd_columns = positions_of_columns_with_type('A');
+    std::vector<int> dano_columns = positions_of_columns_with_type('D');
+    std::vector<std::pair<int,int>> plus_minus_columns = positions_of_plus_minus_columns();
+    bool no_special_columns = phase_columns.empty() && abcd_columns.empty() &&
+                              plus_minus_columns.empty() && dano_columns.empty();
+    bool centric = no_special_columns || gops.is_centrosymmetric();
+    for (size_t n = 0; n < data.size(); n += columns.size()) {
+      Miller hkl = get_hkl(n);
+      if (asu.is_in(hkl))
+        continue;
+      auto result = asu.to_asu(hkl, gops);
+      // cf. impl::move_to_asu() in asudata.hpp
+      set_hkl(n, result.first);
+      if (no_special_columns)
+        continue;
+      int isym = result.second;
+      if (!phase_columns.empty() || !abcd_columns.empty()) {
+        const Op& op = gops.sym_ops[(isym - 1) / 2];
+        double shift = op.phase_shift(hkl);
+        if (shift != 0) {
+          if (isym % 2 == 0)
+            shift = -shift;
+          double shift_deg = deg(shift);
+          for (int col : phase_columns)
+            data[n + col] = float(data[n + col] + shift_deg);
+          for (auto i = abcd_columns.begin(); i+3 < abcd_columns.end(); i += 4) {
+            double sinx = std::sin(shift);
+            double cosx = std::cos(shift);
+            double sin2x = 2 * sinx * cosx;
+            double cos2x = sq(cosx)- sq(sinx);
+            double a = data[n + *(i+0)];
+            double b = data[n + *(i+1)];
+            double c = data[n + *(i+2)];
+            double d = data[n + *(i+3)];
+            // a sin(x+y) + b cos(x+y) = a sin(x) cos(y) - b sin(x) sin(y)
+            //                         + a cos(x) sin(y) + b cos(x) cos(y)
+            data[n + *(i+0)] = float(a * cosx - b * sinx);
+            data[n + *(i+1)] = float(a * sinx + b * cosx);
+            data[n + *(i+2)] = float(c * cos2x - d * sin2x);
+            data[n + *(i+3)] = float(c * sin2x + d * cos2x);
+          }
+        }
+      }
+      if (isym % 2 == 0 && !centric) {
+        for (std::pair<int,int> cols : plus_minus_columns)
+          std::swap(data[n + cols.first], data[n + cols.second]);
+        for (int col : dano_columns)
+          data[n + col] = -data[n + col];
+      }
+    }
+  }
+
   // (for unmerged MTZ only) change HKL according to M/ISYM
   bool switch_to_original_hkl() {
+    if (indices_switched_to_original)
+      return false;
     if (!has_data())
       fail("switch_to_original_hkl(): data not read yet");
     const Column* col = column_with_label("M/ISYM");
@@ -777,11 +901,14 @@ struct Mtz {
       for (int i = 0; i < 3; ++i)
         data[n+i] = static_cast<float>(sign * hkl[i]);
     }
+    indices_switched_to_original = true;
     return true;
   }
 
   // (for unmerged MTZ only) change HKL to ASU equivalent and set ISYM
   bool switch_to_asu_hkl() {
+    if (!indices_switched_to_original)
+      return false;
     if (!has_data())
       fail("switch_to_asu_hkl(): data not read yet");
     const Column* col = column_with_label("M/ISYM");
@@ -796,6 +923,7 @@ struct Mtz {
       float& misym = data[n + misym_idx];
       misym = float(((int)misym & ~0xff) | isym);
     }
+    indices_switched_to_original = false;
     return true;
   }
 
@@ -809,7 +937,7 @@ struct Mtz {
   }
 
   Column& add_column(const std::string& label, char type,
-                     int dataset_id=-1, int pos=-1, bool expand_data=false) {
+                     int dataset_id, int pos, bool expand_data) {
     if (datasets.empty())
       fail("No datasets.");
     if (dataset_id < 0)
@@ -829,21 +957,157 @@ struct Mtz {
     col->parent = this;
     col->idx = pos;
     if (expand_data)
-      expand_data_rows(1);
+      expand_data_rows(1, pos);
     return *col;
   }
 
-  void expand_data_rows(int added) {
-    int old_row_size = (int) columns.size() - added;
-    if ((int) data.size() != old_row_size * nreflections)
+  // helper_functions
+  void check_column(size_t idx, const char* msg) const {
+    if (!has_data())
+      fail(msg, ": data not read yet");
+    if (idx >= columns.size())
+      fail(msg, ": no column with 0-based index ", std::to_string(idx));
+  }
+  void check_trailing_cols(const Column& src_col,
+                           const std::vector<std::string>& trailing_cols) const {
+    assert(src_col.parent == this);
+    if (!has_data())
+      fail("data in source mtz not read yet");
+    if (src_col.idx + trailing_cols.size() >= columns.size())
+      fail("Not enough columns after " + src_col.label);
+    for (size_t i = 0; i < trailing_cols.size(); ++i)
+      if (!trailing_cols[i].empty() &&
+          trailing_cols[i] != columns[src_col.idx + i + 1].label)
+        fail("expected trailing column ", trailing_cols[i], ", found ", src_col.label);
+  }
+  void do_replace_column(size_t dest_idx, const Column& src_col,
+                         const std::vector<std::string>& trailing_cols) {
+    const Mtz* src_mtz = src_col.parent;
+    for (size_t i = 0; i <= trailing_cols.size(); ++i) {
+      Column& dst = columns[dest_idx + i];
+      const Column& src = src_mtz->columns[src_col.idx + i];
+      dst.type = src.type;
+      dst.label = src.label;
+      dst.min_value = src.min_value;
+      dst.max_value = src.max_value;
+      dst.source = src.source;
+      if (src_mtz == this)
+        dst.dataset_id = src.dataset_id;
+    }
+    if (src_mtz == this) {
+      // internal copying
+      for (size_t n = 0; n < data.size(); n += columns.size())
+        for (size_t i = 0; i <= trailing_cols.size(); ++i)
+          data[n + dest_idx + i] = data[n + src_col.idx + i];
+    } else {
+      // external copying - need to match indices
+      std::vector<int> dst_indices = sorted_row_indices();
+      std::vector<int> src_indices = src_mtz->sorted_row_indices();
+      // cf. for_matching_reflections()
+      size_t dst_stride = columns.size();
+      size_t src_stride = src_mtz->columns.size();
+      auto dst = dst_indices.begin();
+      auto src = src_indices.begin();
+      while (dst != dst_indices.end() && src != src_indices.end()) {
+        Miller dst_hkl = get_hkl(*dst * dst_stride);
+        Miller src_hkl = src_mtz->get_hkl(*src * src_stride);
+        if (dst_hkl == src_hkl) {
+          // copy values
+          for (size_t i = 0; i <= trailing_cols.size(); ++i)
+            data[*dst * dst_stride + dest_idx + i] =
+              src_mtz->data[*src * src_stride + src_col.idx + i];
+          ++dst;
+          ++src;
+        } else if (dst_hkl < src_hkl) {
+          ++dst;
+        } else {
+          ++src;
+        }
+      }
+    }
+  }
+
+  // extra_col are columns right after src_col that are also copied.
+  Column& replace_column(size_t dest_idx, const Column& src_col,
+                         const std::vector<std::string>& trailing_cols={}) {
+    src_col.parent->check_trailing_cols(src_col, trailing_cols);
+    check_column(dest_idx + trailing_cols.size(), "replace_column()");
+    do_replace_column(dest_idx, src_col, trailing_cols);
+    return columns[dest_idx];
+  }
+
+  // If dest_idx < 0 - columns are appended at the end
+  // append new column(s), otherwise overwrite existing ones.
+  Column& copy_column(int dest_idx, const Column& src_col,
+                      const std::vector<std::string>& trailing_cols={}) {
+    // check input consistency
+    if (!has_data())
+      fail("copy_column(): data not read yet");
+    src_col.parent->check_trailing_cols(src_col, trailing_cols);
+    // add new columns
+    if (dest_idx < 0)
+      dest_idx = (int) columns.size();
+    // if src_col is from this Mtz it may get invalidated when adding columns
+    int col_idx = -1;
+    if (src_col.parent == this) {
+      col_idx = (int) src_col.idx;
+      if (col_idx >= dest_idx)
+        col_idx += 1 + (int)trailing_cols.size();
+    }
+    for (int i = 0; i <= (int) trailing_cols.size(); ++i)
+      add_column("", ' ', -1, dest_idx + i, false);
+    expand_data_rows(1 + trailing_cols.size(), dest_idx);
+    // copy the data
+    const Column& src_col_now = col_idx < 0 ? src_col : columns[col_idx];
+    do_replace_column(dest_idx, src_col_now, trailing_cols);
+    return columns[dest_idx];
+  }
+
+  void remove_column(size_t idx) {
+    check_column(idx, "remove_column()");
+    columns.erase(columns.begin() + idx);
+    for (size_t i = idx; i < columns.size(); ++i)
+      --columns[i].idx;
+    for (size_t dest = idx, source = idx + 1; source < data.size(); ++source)
+      for (size_t i = 0; i < columns.size() && source < data.size(); ++i)
+        data[dest++] = data[source++];
+    data.resize(columns.size() * nreflections);
+  }
+
+  template <typename Func>
+  void remove_rows_if(Func condition) {
+    if (!has_data())
+      fail("No data.");
+    auto out = data.begin();
+    size_t width = columns.size();
+    for (auto r = data.begin(); r < data.end(); r += width)
+      if (!condition(&*r)) {
+        if (r != out)
+          std::copy(r, r + width, out);
+        out += width;
+      }
+    data.erase(out, data.end());
+    nreflections = int(data.size() / width);
+  }
+
+  void expand_data_rows(size_t added, int pos_=-1) {
+    size_t old_row_size = columns.size() - added;
+    if (data.size() != old_row_size * nreflections)
       fail("Internal error");
     data.resize(columns.size() * nreflections);
+    size_t pos = pos_ == -1 ? old_row_size : (size_t) pos_;
+    if (pos > old_row_size)
+      fail("expand_data_rows(): pos out of range");
+    std::vector<float>::iterator dst = data.end();
     for (int i = nreflections; i-- != 0; ) {
-      for (int j = added; j-- != 0; )
-        data[i * columns.size() + old_row_size + j] = NAN;
-      for (int j = old_row_size; j-- != 0; )
-        data[i * columns.size() + j] = data[i * old_row_size + j];
+      for (size_t j = old_row_size; j-- != pos; )
+        *--dst = data[i * old_row_size + j];
+      for (size_t j = added; j-- != 0; )
+        *--dst = NAN;
+      for (size_t j = pos; j-- != 0; )
+        *--dst = data[i * old_row_size + j];
     }
+    assert(dst == data.begin());
   }
 
   void set_data(const float* new_data, size_t n) {
@@ -857,7 +1121,7 @@ struct Mtz {
   // Function for writing MTZ file
   void write_to_cstream(std::FILE* stream) const;
   void write_to_string(std::string& str) const;
-  template<typename Stream> void write_to_stream(Stream stream) const;
+  template<typename Write> void write_to_stream(Write write) const;
   void write_to_file(const std::string& path) const;
 };
 
@@ -935,10 +1199,17 @@ void Mtz::write_to_stream(Write write) const {
   if (!spacegroup)
     fail("Cannot write Mtz which has no space group");
   char buf[81] = {'M', 'T', 'Z', ' ', '\0'};
-  std::int32_t header_start = (int) columns.size() * nreflections + 21;
+  std::int64_t real_header_start = (int64_t) columns.size() * nreflections + 21;
+  std::int32_t header_start = (int32_t) real_header_start;
+  if (real_header_start > std::numeric_limits<int32_t>::max()) {
+    header_start = -1;
+  } else {
+    real_header_start = 0;
+  }
   std::memcpy(buf + 4, &header_start, 4);
   std::int32_t machst = is_little_endian() ? 0x00004144 : 0x11110000;
   std::memcpy(buf + 8, &machst, 4);
+  std::memcpy(buf + 12, &real_header_start, 8);
   if (write(buf, 80, 1) != 1 ||
       write(data.data(), 4, data.size()) != data.size())
     fail("Writing MTZ file failed");
@@ -980,13 +1251,13 @@ void Mtz::write_to_stream(Write write) const {
   };
   for (const Column& col : columns) {
     auto minmax = calculate_min_max_disregarding_nans(col.begin(), col.end());
+    const char* label = !col.label.empty() ? col.label.c_str() : "_";
     WRITE("COLUMN %-30s %c %17s %17s %4d",
-          col.label.c_str(), col.type,
+          label, col.type,
           format17(minmax[0]).c_str(), format17(minmax[1]).c_str(),
           col.dataset_id);
     if (!col.source.empty())
-      WRITE("COLSRC %-30s %-36s  %4d",
-            col.label.c_str(), col.source.c_str(), col.dataset_id);
+      WRITE("COLSRC %-30s %-36s  %4d", label, col.source.c_str(), col.dataset_id);
   }
   WRITE("NDIF %8zu", datasets.size());
   for (const Dataset& ds : datasets) {
