@@ -7,6 +7,7 @@
 
 #include "model.hpp"
 #include "polyheur.hpp"  // for are_connected
+#include <array>
 #include <ostream>
 
 namespace gemmi {
@@ -14,6 +15,7 @@ namespace gemmi {
 struct PdbWriteOptions {
   bool seqres_records = true;
   bool ssbond_records = true;
+  bool cryst1_record = true;
   bool link_records = true;
   bool cispep_records = true;
   bool ter_records = true;
@@ -28,16 +30,6 @@ void write_minimal_pdb(const Structure& st, std::ostream& os,
                        PdbWriteOptions opt=PdbWriteOptions());
 std::string make_pdb_headers(const Structure& st);
 
-// Name as a string left-padded like in the PDB format:
-// the first two characters make the element name.
-inline std::string padded_atom_name(const Atom& atom) {
-  std::string s;
-  if (atom.element.uname()[1] == '\0' && atom.name.size() < 4)
-    s += ' ';
-  s += atom.name;
-  return s;
-}
-
 } // namespace gemmi
 
 #ifdef GEMMI_WRITE_IMPLEMENTATION
@@ -92,7 +84,7 @@ bool use_hetatm(const Residue& res) {
 }
 
 // works for non-negative values only
-inline char *base36_encode(char* buffer, int width, int value) {
+inline void base36_encode(char* buffer, int width, int value) {
   const char base36[] = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";
   buffer[width] = '\0';
   do {
@@ -101,30 +93,32 @@ inline char *base36_encode(char* buffer, int width, int value) {
   } while (value != 0 && width != 0);
   while (width != 0)
     buffer[--width] = ' ';
-  return buffer;
 }
 
 // based on http://cci.lbl.gov/hybrid_36/
-inline char* encode_serial_in_hybrid36(char* str, int serial) {
+inline std::array<char,8> encode_serial_in_hybrid36(int serial) {
+  std::array<char,8> str;
   assert(serial >= 0);
-  if (serial < 100000) {
-    gstb_sprintf(str, "%5d", serial);
-    return str;
-  }
-  return base36_encode(str, 5, serial - 100000 + 10 * 36 * 36 * 36 * 36);
+  if (serial < 100000)
+    gstb_sprintf(str.data(), "%5d", serial);
+  else
+    base36_encode(str.data(), 5, serial - 100000 + 10 * 36 * 36 * 36 * 36);
+  return str;
 }
 
 // based on http://cci.lbl.gov/hybrid_36/
-inline char* encode_seq_num_in_hybrid36(char* str, int seq_id) {
-  if (seq_id > -1000 && seq_id < 10000) {
-    gstb_sprintf(str, "%4d", seq_id);
-    return str;
-  }
-  return base36_encode(str, 4, seq_id - 10000 + 10 * 36 * 36 * 36);
+inline void encode_seq_num_in_hybrid36(char* str, SeqId::OptionalNum seq_num) {
+  if (*seq_num > -1000 && *seq_num < 10000)
+    gstb_sprintf(str, "%4d", *seq_num);
+  else if (seq_num.has_value())
+    base36_encode(str, 4, *seq_num - 10000 + 10 * 36 * 36 * 36);
+  else
+    std::memcpy(str, "    ", 4);
 }
 
-inline char* write_seq_id(char* str, const SeqId& seqid) {
-  encode_seq_num_in_hybrid36(str, *seqid.num);
+inline std::array<char,8> write_seq_id(const SeqId& seqid) {
+  std::array<char,8> str;
+  encode_seq_num_in_hybrid36(str.data(), *seqid.num);
   str[4] = seqid.icode;
   str[5] = '\0';
   return str;
@@ -272,8 +266,6 @@ inline void write_remarks(const Structure& st, std::ostream& os) {
 inline void write_chain_atoms(const Chain& chain, std::ostream& os,
                               int& serial, PdbWriteOptions opt) {
   char buf[88];
-  char buf8[8];
-  char buf8a[8];
   buf[0] = '\0';
   if (chain.name.length() > 2)
     fail("long chain name: " + chain.name);
@@ -304,12 +296,12 @@ inline void write_chain_atoms(const Chain& chain, std::ostream& os,
             "%2s%5s   %8.3f%8.3f%8.3f"
             "%6.2f%6.2f      %-4.4s%2s%c%c",
             as_het ? "HETATM" : "ATOM",
-            impl::encode_serial_in_hybrid36(buf8, ++serial),
-            padded_atom_name(a).c_str(),
+            impl::encode_serial_in_hybrid36(++serial).data(),
+            a.padded_name().c_str(),
             a.altloc ? std::toupper(a.altloc) : ' ',
             res.name.c_str(),
             chain.name.c_str(),
-            impl::write_seq_id(buf8a, res.seqid),
+            impl::write_seq_id(res.seqid).data(),
             // We want to avoid negative zero and round the numbers up
             // if they originally had one digit more and that digit was 5.
             a.pos.x > -5e-4 && a.pos.x < 0 ? 0 : a.pos.x + 1e-10,
@@ -337,6 +329,7 @@ inline void write_chain_atoms(const Chain& chain, std::ostream& os,
                     a.aniso.u33*1e4 + eps, a.aniso.u12*1e4 + eps,
                     a.aniso.u13*1e4 + eps, a.aniso.u23*1e4 + eps);
         buf[28+42] = ' ';
+        buf[80] = '\n';
         os.write(buf, 81);
       }
     }
@@ -349,9 +342,10 @@ inline void write_chain_atoms(const Chain& chain, std::ostream& os,
         // re-using part of the buffer in the middle, e.g.:
         // TER    4153      LYS B 286
         gf_snprintf(buf, 82, "TER   %5s",
-                    impl::encode_serial_in_hybrid36(buf8, ++serial));
+                    impl::encode_serial_in_hybrid36(++serial).data());
         std::memset(buf+11, ' ', 6);
         std::memset(buf+28, ' ', 52);
+        buf[80] = '\n';
         os.write(buf, 81);
       } else {
         WRITE("%-80s", "TER");
@@ -445,6 +439,7 @@ inline void write_header(const Structure& st, std::ostream& os,
     for (size_t i = 0; i != entity_list.size(); ++i)
       if (const Entity* entity = entity_list[i]) {
         const Chain& ch = st.models[0].chains[i];
+        const char* dbref_entry_id = entry_id.size() <= 4 ? entry_id.c_str() : "";
         for (const Entity::DbRef& dbref : entity->dbrefs) {
           bool short_record = *dbref.db_end.num < 100000 &&
                               dbref.accession_code.size() < 9 &&
@@ -456,14 +451,15 @@ inline void write_header(const Structure& st, std::ostream& os,
               begin = polymer.label_seq_id_to_auth(dbref.label_seq_begin);
               end = polymer.label_seq_id_to_auth(dbref.label_seq_end);
             }
-          char buf8[8];
-          char buf8a[8];
           gf_snprintf(buf, 82, "DBREF  %4s%2s %5s %5s %-6s  ",
-                      entry_id.c_str(), ch.name.c_str(),
-                      impl::write_seq_id(buf8, begin),
-                      impl::write_seq_id(buf8a, end),
+                      dbref_entry_id, ch.name.c_str(),
+                      impl::write_seq_id(begin).data(),
+                      impl::write_seq_id(end).data(),
                       dbref.db_name.c_str());
-          if (!(dbref.db_name == "PDB" && dbref.id_code == entry_id)) {
+          if (dbref.db_name == "PDB" && dbref.id_code == entry_id) {
+            // PDB uses self-reference for fragments that don't have real
+            // reference. No idea why. In such case the same begin/end is used.
+          } else {
             begin = dbref.db_begin;
             end = dbref.db_end;
           }
@@ -472,14 +468,15 @@ inline void write_header(const Structure& st, std::ostream& os,
                         dbref.accession_code.c_str(), dbref.id_code.c_str(),
                         *begin.num, begin.icode, *end.num, end.icode);
           } else {
-            buf[5] = '1';
+            buf[5] = '1';  // -> DBREF1
             gf_snprintf(buf+33, 82-33, "              %-33s\n",
                         dbref.id_code.c_str());
           }
+          buf[80] = '\n';
           os.write(buf, 81);
           if (!short_record)
             WRITE("DBREF2 %4s%2s     %-22s     %10d  %10d             ",
-                  entry_id.c_str(), ch.name.c_str(),
+                  dbref_entry_id, ch.name.c_str(),
                   dbref.accession_code.c_str(), *begin.num, *end.num);
         }
       }
@@ -503,14 +500,14 @@ inline void write_header(const Structure& st, std::ostream& os,
             col = 0;
           }
         }
-        if (col != 0)
+        if (col != 0) {
+          buf[80] = '\n';
           os.write(buf, 81);
+        }
       }
   }
 
   if (!st.helices.empty()) {
-    char buf8[8];
-    char buf8a[8];
     int counter = 0;
     for (const Helix& helix : st.helices) {
       if (++counter == 10000)
@@ -520,18 +517,18 @@ inline void write_header(const Structure& st, std::ostream& os,
       gf_snprintf(buf, 82, "HELIX %4d%4d %3s%2s %5s %3s%2s %5s%2d %35d    \n",
             counter, counter,
             helix.start.res_id.name.c_str(), helix.start.chain_name.c_str(),
-            write_seq_id(buf8, helix.start.res_id.seqid),
+            write_seq_id(helix.start.res_id.seqid).data(),
             helix.end.res_id.name.c_str(), helix.end.chain_name.c_str(),
-            write_seq_id(buf8a, helix.end.res_id.seqid),
+            write_seq_id(helix.end.res_id.seqid).data(),
             (int) helix.pdb_helix_class, helix.length);
       if (helix.length < 0) // make 72-76 blank if the length is not given
         std::memset(buf+71, ' ', 5);
+      buf[80] = '\n';
       os.write(buf, 81);
     }
   }
 
   if (!st.sheets.empty()) {
-    char buf8a[8], buf8b[8], buf8c[8], buf8d[8];
     for (const Sheet& sheet : st.sheets) {
       int strand_counter = 0;
       for (const Sheet::Strand& strand : sheet.strands) {
@@ -542,22 +539,20 @@ inline void write_header(const Structure& st, std::ostream& os,
               " %-3s%3s%2s%5s          ",
               ++strand_counter, sheet.name.c_str(), sheet.strands.size(),
               strand.start.res_id.name.c_str(), strand.start.chain_name.c_str(),
-              write_seq_id(buf8a, strand.start.res_id.seqid),
+              write_seq_id(strand.start.res_id.seqid).data(),
               strand.end.res_id.name.c_str(), strand.end.chain_name.c_str(),
-              write_seq_id(buf8b, strand.end.res_id.seqid), strand.sense,
+              write_seq_id(strand.end.res_id.seqid).data(), strand.sense,
               a2.atom_name.c_str(), a2.res_id.name.c_str(),
               a2.chain_name.c_str(),
-              a2.res_id.seqid.num ? write_seq_id(buf8c, a2.res_id.seqid) : "",
+              a2.res_id.seqid.num ? write_seq_id(a2.res_id.seqid).data() : "",
               a1.atom_name.c_str(), a1.res_id.name.c_str(),
               a1.chain_name.c_str(),
-              a1.res_id.seqid.num ? write_seq_id(buf8d, a1.res_id.seqid) : "");
+              a1.res_id.seqid.num ? write_seq_id(a1.res_id.seqid).data() : "");
       }
     }
   }
 
   if (!st.models.empty()) {
-    char buf8[8];
-    char buf8a[8];
     // SSBOND  (note: uses only the first model and primary conformation)
     if (opt.ssbond_records) {
       int counter = 0;
@@ -567,17 +562,17 @@ inline void write_header(const Structure& st, std::ostream& os,
           const_CRA cra2 = st.models[0].find_cra(con.partner2, true);
           if (!cra1.atom || !cra2.atom)
             continue;
-          SymImage im = st.cell.find_nearest_image(cra1.atom->pos,
-                                                   cra2.atom->pos, con.asu);
+          NearestImage im = st.cell.find_nearest_image(cra1.atom->pos,
+                                                       cra2.atom->pos, con.asu);
           if (++counter == 10000)
             counter = 0;
           WRITE("SSBOND%4d %3s%2s %5s %5s%2s %5s %28s %6s %5.2f  ",
              counter,
              cra1.residue->name.c_str(), cra1.chain->name.c_str(),
-             write_seq_id(buf8, cra1.residue->seqid),
+             write_seq_id(cra1.residue->seqid).data(),
              cra2.residue->name.c_str(), cra2.chain->name.c_str(),
-             write_seq_id(buf8a, cra2.residue->seqid),
-             "1555", im.pdb_symbol(false).c_str(), im.dist());
+             write_seq_id(cra2.residue->seqid).data(),
+             "1555", im.symmetry_code(false).c_str(), im.dist());
         }
     }
 
@@ -591,12 +586,12 @@ inline void write_header(const Structure& st, std::ostream& os,
           // In special cases (LINKR gap) atoms are not there.
           if (!cra1.residue || !cra2.residue)
             continue;
-          std::string im_pdb_symbol = "", im_dist_str = "";
+          std::string im_pdb_symbol, im_dist_str;
           bool im_same_asu = true;
           if (cra1.atom && cra2.atom) {
-            SymImage im = st.cell.find_nearest_image(cra1.atom->pos,
-                                                     cra2.atom->pos, con.asu);
-            im_pdb_symbol = im.pdb_symbol(false);
+            NearestImage im = st.cell.find_nearest_image(cra1.atom->pos,
+                                                         cra2.atom->pos, con.asu);
+            im_pdb_symbol = im.symmetry_code(false);
             im_dist_str = to_str_prec<2>(im.dist());
             im_same_asu = im.same_asu();
           }
@@ -607,16 +602,16 @@ inline void write_header(const Structure& st, std::ostream& os,
           // except for LINKR (Refmac variant of LINK).
           gf_snprintf(buf, 82, "LINK        %-4s%c%3s%2s%5s   "
                 "            %-4s%c%3s%2s%5s  %6s %6s %5s  \n",
-                cra1.atom ? padded_atom_name(*cra1.atom).c_str() : "",
+                cra1.atom ? cra1.atom->padded_name().c_str() : "",
                 cra1.atom && cra1.atom->altloc ? std::toupper(cra1.atom->altloc) : ' ',
                 cra1.residue->name.c_str(),
                 con.partner1.chain_name.c_str(),
-                write_seq_id(buf8, cra1.residue->seqid),
-                cra2.atom ? padded_atom_name(*cra2.atom).c_str() : "",
+                write_seq_id(cra1.residue->seqid).data(),
+                cra2.atom ? cra2.atom->padded_name().c_str() : "",
                 cra2.atom && cra2.atom->altloc ? std::toupper(cra2.atom->altloc) : ' ',
                 cra2.residue->name.c_str(),
                 con.partner2.chain_name.c_str(),
-                write_seq_id(buf8a, cra2.residue->seqid),
+                write_seq_id(cra2.residue->seqid).data(),
                 "1555", im_pdb_symbol.c_str(), im_dist_str.c_str());
           if (opt.use_linkr && !con.link_id.empty()) {
             buf[4] = 'R';  // LINK -> LINKR
@@ -625,6 +620,7 @@ inline void write_header(const Structure& st, std::ostream& os,
             // overwrite distance with link_id
             gf_snprintf(buf+72, 82-72, "%-8s\n", con.link_id.c_str());
           }
+          buf[80] = '\n';
           os.write(buf, 81);
         }
 
@@ -644,9 +640,9 @@ inline void write_header(const Structure& st, std::ostream& os,
                 WRITE("CISPEP%4d %3s%2s %5s   %3s%2s %5s %9s %12.2f %20s",
                       counter,
                       res.name.c_str(), chain.name.c_str(),
-                      write_seq_id(buf8, res.seqid),
+                      write_seq_id(res.seqid).data(),
                       next->name.c_str(), chain.name.c_str(),
-                      write_seq_id(buf8a, next->seqid),
+                      write_seq_id(next->seqid).data(),
                       st.models.size() > 1 ? model.name.c_str() : "0",
                       deg(calculate_omega(res, *next)),
                       "");
@@ -655,7 +651,8 @@ inline void write_header(const Structure& st, std::ostream& os,
     }
   }
 
-  write_cryst1(st, os);
+  if(opt.cryst1_record)
+    write_cryst1(st, os);
   if (st.has_origx && !st.origx.is_identity()) {
     for (int i = 0; i < 3; ++i)
       WRITE("ORIGX%d %13.6f%10.6f%10.6f %14.5f %24s", i+1,
