@@ -2,9 +2,11 @@
 
 #include "gemmi/model.hpp"
 #include "gemmi/calculate.hpp"  // for calculate_mass, count_atom_sites
+#include "gemmi/modify.hpp"     // for remove_alternative_conformations
 #include "gemmi/polyheur.hpp"   // for one_letter_code, trim_to_alanine
 #include "gemmi/assembly.hpp"   // for expand_ncs, HowToNameCopiedChain
-#include "gemmi/tostr.hpp"
+#include "gemmi/select.hpp"     // for Selection
+#include "tostr.hpp"
 
 #include "common.h"
 #include <pybind11/stl.h>
@@ -34,6 +36,11 @@ template<typename P, typename C> C& get_child(P& parent, int index) {
   return children[normalize_index(index, children)];
 }
 
+template<typename P, typename C> void set_child(P& parent, int index, C& child) {
+  auto& children = parent.children();
+  children[normalize_index(index, children)] = child;
+}
+
 template<typename P> void remove_child(P& parent, int index) {
   auto& children = parent.children();
   children.erase(children.begin() + normalize_index(index, children));
@@ -53,6 +60,7 @@ void add_mol(py::module& m) {
   py::class_<Model> pyModel(m, "Model");
   py::class_<ResidueSpan> pyResidueSpan(m, "ResidueSpan");
   py::class_<ResidueGroup, ResidueSpan> pyResidueGroup(m, "ResidueGroup");
+  py::class_<Selection> pySelection(m, "Selection");
 
   py::enum_<HowToNameCopiedChain>(m, "HowToNameCopiedChain")
     .value("Short", HowToNameCopiedChain::Short)
@@ -68,7 +76,7 @@ void add_mol(py::module& m) {
 
   py::enum_<CoorFormat>(m, "CoorFormat")
     .value("Unknown", CoorFormat::Unknown)
-    .value("UnknownAny", CoorFormat::UnknownAny)
+    .value("Detect", CoorFormat::Detect)
     .value("Pdb", CoorFormat::Pdb)
     .value("Mmcif", CoorFormat::Mmcif)
     .value("Mmjson", CoorFormat::Mmjson)
@@ -76,6 +84,22 @@ void add_mol(py::module& m) {
 
   py::bind_map<info_map_type>(m, "InfoMap");
 
+  py::class_<CRA>(m, "CRA")
+    .def_readonly("chain", &CRA::chain)
+    .def_readonly("residue", &CRA::residue)
+    .def_readonly("atom", &CRA::atom)
+    .def("atom_matches", [](const CRA& self, const AtomAddress& addr) {
+        return atom_matches(self, addr);
+    })
+    .def("__str__", [](const CRA& self) { return atom_str(self); })
+    .def("__repr__", [](const CRA& self) {
+        return tostr("<gemmi.CRA ", atom_str(self), '>');
+    });
+
+  py::class_<CraProxy>(m, "CraGenerator")
+    .def("__iter__", [](CraProxy& self) { return py::make_iterator(self); },
+         py::keep_alive<0, 1>());
+
   py::class_<Structure> structure(m, "Structure");
   structure
     .def(py::init<>())
@@ -84,9 +108,14 @@ void add_mol(py::module& m) {
     .def_readwrite("spacegroup_hm", &Structure::spacegroup_hm)
     .def_readwrite("ncs", &Structure::ncs)
     .def_readwrite("resolution", &Structure::resolution)
+    .def_readwrite("input_format", &Structure::input_format)
     .def_readwrite("entities", &Structure::entities)
     .def_readwrite("connections", &Structure::connections)
+    .def_readwrite("helices", &Structure::helices)
+    .def_readwrite("sheets", &Structure::sheets)
     .def_readwrite("assemblies", &Structure::assemblies)
+    .def_readwrite("meta", &Structure::meta)
+    .def_readwrite("has_d_fraction", &Structure::has_d_fraction)
     .def_readwrite("info", &Structure::info)
     .def_readwrite("raw_remarks", &Structure::raw_remarks)
     .def("find_spacegroup", &Structure::find_spacegroup)
@@ -108,6 +137,10 @@ void add_mol(py::module& m) {
     .def("__delitem__", &remove_child<Structure>, py::arg("index"))
     .def("__delitem__", &remove_children<Structure>)
     .def("__delitem__", &Structure::remove_model, py::arg("name"))
+    .def("__setitem__", &set_child<Structure, Model>)
+    .def("find_connection_by_cra", [](Structure& st, CRA cra1, CRA cra2) {
+        return st.find_connection_by_cra(cra1, cra2);
+    }, py::return_value_policy::reference_internal)
     .def("find_connection", &Structure::find_connection,
          py::arg("partner1"), py::arg("partner2"),
          py::return_value_policy::reference_internal)
@@ -123,19 +156,26 @@ void add_mol(py::module& m) {
     .def("setup_cell_images", &Structure::setup_cell_images)
     .def("add_entity_types", (void (*)(Structure&, bool)) &add_entity_types,
          py::arg("overwrite")=false)
-    .def("assign_subchains", (void (*)(Structure&, bool)) &assign_subchains,
-         py::arg("overwrite")=false)
+    .def("assign_subchains", &assign_subchains,
+         py::arg("force")=false, py::arg("fail_if_unknown")=true)
     .def("ensure_entities", &ensure_entities)
     .def("deduplicate_entities", &deduplicate_entities)
     .def("setup_entities", &setup_entities)
+    .def("assign_cis_flags", assign_cis_flags<Structure>)
     .def("remove_alternative_conformations",
          remove_alternative_conformations<Structure>)
     .def("remove_hydrogens", remove_hydrogens<Structure>)
     .def("remove_waters", remove_waters<Structure>)
-    .def("remove_ligands_and_waters",
-         (void (*)(Structure&)) &remove_ligands_and_waters)
+    .def("remove_ligands_and_waters", remove_ligands_and_waters<Structure>)
+    .def("expand_hd_mixture", &expand_hd_mixture)
+    .def("collapse_hd_mixture", &collapse_hd_mixture)
+    .def("assign_serial_numbers", (void (*)(Structure&)) &assign_serial_numbers)
     .def("shorten_chain_names", &shorten_chain_names)
     .def("expand_ncs", &expand_ncs, py::arg("how"))
+    .def("transform_to_assembly",
+         [](Structure& st, const std::string& assembly_name, HowToNameCopiedChain how) {
+        return transform_to_assembly(st, assembly_name, how, nullptr);
+    }, py::arg("assembly_name"), py::arg("how"))
     .def("calculate_box", &calculate_box, py::arg("margin")=0.)
     .def("calculate_fractional_box", &calculate_fractional_box, py::arg("margin")=0.)
     .def("clone", [](const Structure& self) { return new Structure(self); })
@@ -146,22 +186,6 @@ void add_mol(py::module& m) {
     add_assign_label_seq_id(structure);
     add_write(m, structure);
 
-  py::class_<CRA>(m, "CRA")
-    .def_readonly("chain", &CRA::chain)
-    .def_readonly("residue", &CRA::residue)
-    .def_readonly("atom", &CRA::atom)
-    .def("atom_matches", [](const CRA& self, const AtomAddress& addr) {
-        return atom_matches(self, addr);
-    })
-    .def("__str__", [](const CRA& self) { return atom_str(self); })
-    .def("__repr__", [](const CRA& self) {
-        return tostr("<gemmi.CRA ", atom_str(self), '>');
-    });
-
-  py::class_<CraProxy>(m, "CraGenerator")
-    .def("__iter__", [](CraProxy& self) { return py::make_iterator(self); },
-         py::keep_alive<0, 1>());
-
   pyModel
     .def(py::init<std::string>())
     .def_readwrite("name", &Model::name)
@@ -211,17 +235,20 @@ void add_mol(py::module& m) {
     .def("__delitem__", &Model::remove_chain, py::arg("name"))
     .def("__delitem__", remove_child<Model>, py::arg("index"))
     .def("__delitem__", remove_children<Model>)
-    .def("count_atom_sites", &count_atom_sites<Model>)
+    .def("remove_alternative_conformations", remove_alternative_conformations<Model>)
+    .def("remove_hydrogens", remove_hydrogens<Model>)
+    .def("remove_waters", remove_waters<Model>)
+    .def("remove_ligands_and_waters", remove_ligands_and_waters<Model>)
+    .def("has_hydrogen", &has_hydrogen<Model>)
+    .def("count_atom_sites", &count_atom_sites<Model>, py::arg("sel")=nullptr)
+    // deprecated, use has_hydrogen() or count_atom_sites(Selection('[H,D]'))
     .def("count_hydrogen_sites", &count_hydrogen_sites<Model>)
-    .def("count_occupancies", &count_occupancies<Model>)
+    .def("count_occupancies", &count_occupancies<Model>, py::arg("sel")=nullptr)
     .def("calculate_mass", &calculate_mass<Model>)
     .def("calculate_center_of_mass", [](const Model& self) {
         return calculate_center_of_mass(self).get();
     })
-    .def("transform", [](Model& self, const Transform& tr) {
-        for (CRA cra : self.all())
-          transform_atom(*cra.atom, tr);
-    }, py::arg("tr"))
+    .def("transform_pos_and_adp", transform_pos_and_adp<Model>, py::arg("tr"))
     .def("split_chains_by_segments", &split_chains_by_segments)
     .def("clone", [](const Model& self) { return new Model(self); })
     .def("__repr__", [](const Model& self) {
@@ -280,15 +307,15 @@ void add_mol(py::module& m) {
          py::keep_alive<0, 1>())
     .def("get_subchain", (ResidueSpan (Chain::*)(const std::string&)) &Chain::get_subchain,
          py::keep_alive<0, 1>())
-    .def("has_subchains_assigned", &has_subchains_assigned)
+    .def("has_entity_types_and_subchains", &has_entity_types_and_subchains)
     .def("previous_residue", &Chain::previous_residue,
          py::return_value_policy::reference_internal)
     .def("next_residue", &Chain::next_residue,
          py::return_value_policy::reference_internal)
     .def("append_residues", &Chain::append_residues,
          py::arg("new_residues"), py::arg("min_sep")=0)
-    .def("count_atom_sites", &count_atom_sites<Chain>)
-    .def("count_occupancies", &count_occupancies<Chain>)
+    .def("count_atom_sites", &count_atom_sites<Chain>, py::arg("sel")=nullptr)
+    .def("count_occupancies", &count_occupancies<Chain>, py::arg("sel")=nullptr)
     .def("calculate_mass", &calculate_mass<Chain>)
     .def("calculate_center_of_mass", [](const Chain& self) {
           return calculate_center_of_mass(self).get();
@@ -335,9 +362,17 @@ void add_mol(py::module& m) {
     .def("make_one_letter_sequence", [](const ResidueSpan& span) {
         return make_one_letter_sequence(span);
     })
+    .def("transform_pos_and_adp", transform_pos_and_adp<ResidueSpan>)
     .def("__repr__", [](const ResidueSpan& self) {
         int N = (int) self.size();
-        std::string r = "<gemmi.ResidueSpan of " + std::to_string(N) + ": [";
+        std::string r = "<gemmi.ResidueSpan of " + std::to_string(N) + ": ";
+        if (N > 0) {
+          r += self.front().subchain;
+          if (self.back().subchain != self.front().subchain)
+            r += " - " + self.back().subchain;
+          r += ' ';
+        }
+        r += '[';
         for (int i = 0; i < (N < 5 ? N : 3); ++i) {
           if (i != 0) r += ' ';
           r += self[i].str();
@@ -380,10 +415,16 @@ void add_mol(py::module& m) {
   pyResidue
     .def(py::init<>())
     .def_readwrite("subchain", &Residue::subchain)
+    .def_readonly("entity_id", &Residue::entity_id)
+    .def_readwrite("label_seq", &Residue::label_seq)
     .def_readwrite("entity_type", &Residue::entity_type)
     .def_readwrite("het_flag", &Residue::het_flag)
     .def_readwrite("flag", &Residue::flag)
-    .def_readwrite("label_seq", &Residue::label_seq)
+    .def_property_readonly("sifts_unp", [](const Residue& self) {
+        return py::make_tuple(self.sifts_unp.res,
+                              self.sifts_unp.acc_index,
+                              self.sifts_unp.num);
+    })
     .def("__len__", [](const Residue& res) { return res.atoms.size(); })
     .def("__contains__", [](const Residue& res, const std::string& name) {
         return res.find_atom(name, '*') != nullptr;
@@ -440,6 +481,7 @@ void add_mol(py::module& m) {
     .def_readwrite("b_iso", &Atom::b_iso)
     .def_readwrite("aniso", &Atom::aniso)
     .def_readwrite("serial", &Atom::serial)
+    .def_readwrite("fraction", &Atom::fraction)
     .def_readwrite("calc_flag", &Atom::calc_flag)
     .def_readwrite("flag", &Atom::flag)
     .def_readwrite("tls_group_id", &Atom::tls_group_id)
@@ -480,4 +522,54 @@ void add_mol(py::module& m) {
                             HowToNameCopiedChain how) {
         return make_assembly(assembly, model, how, nullptr);
   });
+
+  // select.hpp
+  py::class_<FilterProxy<Selection, Model>> pySelectionModelsProxy(m, "SelectionModelsProxy");
+  py::class_<FilterProxy<Selection, Chain>> pySelectionChainsProxy(m, "SelectionChainsProxy");
+  py::class_<FilterProxy<Selection, Residue>> pySelectionResiduesProxy(m, "SelectionResiduesProxy");
+  py::class_<FilterProxy<Selection, Atom>> pySelectionAtomsProxy(m, "SelectionAtomsProxy");
+
+  pySelection
+    .def(py::init<>())
+    .def(py::init<const std::string&>())
+    .def("models", &Selection::models)
+    .def("chains", &Selection::chains)
+    .def("residues", &Selection::residues)
+    .def("atoms", &Selection::atoms)
+    .def("first_in_model", &Selection::first_in_model,
+         py::keep_alive<1, 2>())
+    .def("first", &Selection::first, py::return_value_policy::reference,
+         py::keep_alive<1, 2>())
+    .def("str", &Selection::str)
+    .def("set_residue_flags", &Selection::set_residue_flags)
+    .def("set_atom_flags", &Selection::set_atom_flags)
+    .def("copy_model_selection", &Selection::copy_selection<Model>)
+    .def("copy_structure_selection", &Selection::copy_selection<Structure>)
+    .def("remove_selected", &Selection::remove_selected<Structure>)
+    .def("remove_selected", &Selection::remove_selected<Model>)
+    .def("remove_not_selected", &Selection::remove_not_selected<Structure>)
+    .def("remove_not_selected", &Selection::remove_not_selected<Model>)
+    .def("__repr__", [](const Selection& self) {
+        return "<gemmi.Selection CID: " + self.str() + ">";
+    });
+
+  pySelectionModelsProxy
+    .def("__iter__", [](FilterProxy<Selection, Model>& self) {
+        return py::make_iterator(self);
+    }, py::keep_alive<0, 1>());
+
+  pySelectionChainsProxy
+    .def("__iter__", [](FilterProxy<Selection, Chain>& self) {
+        return py::make_iterator(self);
+    }, py::keep_alive<0, 1>());
+
+  pySelectionResiduesProxy
+    .def("__iter__", [](FilterProxy<Selection, Residue>& self) {
+        return py::make_iterator(self);
+    }, py::keep_alive<0, 1>());
+
+  pySelectionAtomsProxy
+    .def("__iter__", [](FilterProxy<Selection, Atom>& self) {
+        return py::make_iterator(self);
+    }, py::keep_alive<0, 1>());
 }
