@@ -1,9 +1,10 @@
 // Copyright 2019 Global Phasing Ltd.
 
 #include "gemmi/mtz.hpp"
+#include "gemmi/reindex.hpp"  // for reindex_mtz
 #include "gemmi/fourier.hpp"
 #include "gemmi/gz.hpp"
-#include "gemmi/tostr.hpp"
+#include "tostr.hpp"
 
 #include "common.h"
 #include <pybind11/stl.h>
@@ -83,14 +84,23 @@ void add_mtz(py::module& m) {
                              {nrow, ncol}, // dimensions
                              {4 * ncol, 4});  // strides
     })
+    .def_property_readonly("array", [](const Mtz& self) {
+      int nrow = self.has_data() ? self.nreflections : 0;
+      int ncol = (int) self.columns.size();
+      return py::array_t<float>({nrow, ncol}, {4 * ncol, 4},
+                                self.data.data(), py::cast(self));
+    }, py::return_value_policy::reference_internal)
     .def_readwrite("title", &Mtz::title)
     .def_readwrite("nreflections", &Mtz::nreflections)
+    .def_readwrite("sort_order", &Mtz::sort_order)
     .def_readwrite("min_1_d2", &Mtz::min_1_d2)
     .def_readwrite("max_1_d2", &Mtz::max_1_d2)
     .def_readwrite("valm", &Mtz::valm)
-    .def_readwrite("nsymop", &Mtz::nsymop)
+    .def_readonly("nsymop", &Mtz::nsymop)
     .def_readwrite("cell", &Mtz::cell)
     .def_readwrite("spacegroup", &Mtz::spacegroup)
+    .def_readonly("spacegroup_name", &Mtz::spacegroup_name)
+    .def_readonly("spacegroup_number", &Mtz::spacegroup_number)
     .def_readwrite("datasets", &Mtz::datasets)
     .def_readwrite("columns", &Mtz::columns)
     .def_readwrite("batches", &Mtz::batches)
@@ -197,6 +207,15 @@ void add_mtz(py::module& m) {
     .def("add_column", &Mtz::add_column, py::arg("label"), py::arg("type"),
          py::arg("dataset_id")=-1, py::arg("pos")=-1, py::arg("expand_data")=true,
          py::return_value_policy::reference_internal)
+    .def("replace_column", &Mtz::replace_column,
+         py::arg("dest_idx"), py::arg("src_col"),
+         py::arg("trailing_cols")=std::vector<std::string>(),
+         py::return_value_policy::reference_internal)
+    .def("copy_column", &Mtz::copy_column,
+         py::arg("dest_idx"), py::arg("src_col"),
+         py::arg("trailing_cols")=std::vector<std::string>(),
+         py::return_value_policy::reference_internal)
+    .def("remove_column", &Mtz::remove_column, py::arg("index"))
     .def("set_data", [](Mtz& self, const AsuData<std::complex<float>>& asu_data) {
          if (self.columns.size() != 5)
            fail("Mtz.set_data(): Mtz must have 5 columns to put H,K,L,F,Phi.");
@@ -232,9 +251,16 @@ void add_mtz(py::module& m) {
              self.data[row*ncol+col] = r(row, col);
     }, py::arg("array"))
     .def("update_reso", &Mtz::update_reso)
+    .def("sort", &Mtz::sort, py::arg("use_first")=3)
+    .def("ensure_asu", &Mtz::ensure_asu, py::arg("tnt_asu")=false)
     .def("switch_to_original_hkl", &Mtz::switch_to_original_hkl)
     .def("switch_to_asu_hkl", &Mtz::switch_to_asu_hkl)
     .def("write_to_file", &Mtz::write_to_file, py::arg("path"))
+    .def("reindex", [](Mtz& self, const Op& op) {
+        std::ostringstream out;
+        reindex_mtz(self, op, &out);
+        return out.str();
+    }, py::arg("op"))
     .def("__repr__", [](const Mtz& self) {
         return tostr("<gemmi.Mtz with ", self.columns.size(), " columns, ",
                      self.nreflections, " reflections>");
