@@ -5,8 +5,9 @@
 #include "gemmi/cif.hpp"
 #include "gemmi/json.hpp"
 #include "gemmi/smcif.hpp"         // for make_small_structure_from_block
+#include "gemmi/small.hpp"         // for SmallStructure
+#include "gemmi/interop.hpp"       // for atom_to_site, mx_to_sx_structure
 #include "gemmi/chemcomp_xyz.hpp"
-#include "gemmi/remarks.hpp"
 
 #define GEMMI_READ_CIF_IMPLEMENTATION
 #include "gemmi/read_cif.hpp" // for read_cif_gz, read_mmjson_gz
@@ -14,10 +15,14 @@
 #include "gemmi/read_coor.hpp"  // for read_structure_gz
 
 #include "common.h"
+#include <pybind11/stl.h>
+#include <pybind11/stl_bind.h>
 
 namespace py = pybind11;
 using namespace gemmi;
 
+PYBIND11_MAKE_OPAQUE(std::vector<SmallStructure::Site>)
+
 void add_cif_read(py::module& cif) {
   cif.def("read_file", &cif::read_file, py::arg("filename"),
           "Reads a CIF file copying data into Document.");
@@ -43,15 +48,14 @@ void add_cif_read(py::module& cif) {
 
 void add_read_structure(py::module& m) {
   m.def("read_structure", [](const std::string& path, bool merge,
-                             CoorFormat format) {
-          Structure* st = new Structure(read_structure_gz(path, format));
-          if (!st->raw_remarks.empty())
-            read_metadata_from_remarks(*st);
+                             CoorFormat format, cif::Document* save_doc) {
+          Structure* st = new Structure(read_structure_gz(path, format, save_doc));
           if (merge)
             st->merge_chain_parts();
           return st;
         }, py::arg("path"), py::arg("merge_chain_parts")=true,
            py::arg("format")=CoorFormat::Unknown,
+           py::arg("save_doc")=nullptr,
         "Reads a coordinate file into Structure.");
   m.def("make_structure_from_block", &make_structure_from_block,
         py::arg("block"), "Takes mmCIF block and returns Structure.");
@@ -60,9 +64,7 @@ void add_read_structure(py::module& m) {
           PdbReadOptions options;
           options.max_line_length = max_line_length;
           options.split_chain_on_ter = split_chain_on_ter;
-          Structure* st = new Structure(read_pdb_string(s, "string", options));
-          read_metadata_from_remarks(*st);
-          return st;
+          return new Structure(read_pdb_string(s, "string", options));
         }, py::arg("s"), py::arg("max_line_length")=0,
            py::arg("split_chain_on_ter")=false, "Reads a string as PDB file.");
   m.def("read_pdb", [](const std::string& path, int max_line_length,
@@ -70,9 +72,7 @@ void add_read_structure(py::module& m) {
           PdbReadOptions options;
           options.max_line_length = max_line_length;
           options.split_chain_on_ter = split_chain_on_ter;
-          Structure* st = new Structure(read_pdb_gz(path, options));
-          read_metadata_from_remarks(*st);
-          return st;
+          return new Structure(read_pdb_gz(path, options));
         }, py::arg("filename"), py::arg("max_line_length")=0,
            py::arg("split_chain_on_ter")=false);
 
@@ -96,6 +96,64 @@ void add_read_structure(py::module& m) {
         "Returns uncompressed size of a .gz file (not always reliable)");
 }
 
+void add_small(py::module& m) {
+  using gemmi::SmallStructure;
+  py::class_<SmallStructure> small_structure(m, "SmallStructure");
+  py::class_<SmallStructure::Site>(small_structure, "Site")
+    .def(py::init<>())
+    .def(py::init(&gemmi::atom_to_site))
+    .def_readwrite("label", &SmallStructure::Site::label)
+    .def_readwrite("type_symbol", &SmallStructure::Site::type_symbol)
+    .def_readwrite("fract", &SmallStructure::Site::fract)
+    .def_readwrite("occ", &SmallStructure::Site::occ)
+    .def_readwrite("u_iso", &SmallStructure::Site::u_iso)
+    .def_readwrite("element", &SmallStructure::Site::element)
+    .def_readwrite("charge", &SmallStructure::Site::charge)
+    .def_readwrite("disorder_group", &SmallStructure::Site::disorder_group)
+    .def_readwrite("aniso", &SmallStructure::Site::aniso)
+    .def("orth", &SmallStructure::Site::orth)
+    .def("__repr__", [](const SmallStructure::Site& self) {
+        return "<gemmi.SmallStructure.Site " + self.label + ">";
+    });
+  py::bind_vector<std::vector<SmallStructure::Site>>(small_structure, "SiteList");
+
+  using AtomType = SmallStructure::AtomType;
+  py::class_<AtomType>(small_structure, "AtomType")
+    .def_readonly("symbol", &AtomType::symbol)
+    .def_readonly("element", &AtomType::element)
+    .def_readwrite("dispersion_real", &AtomType::dispersion_real)
+    .def_readwrite("dispersion_imag", &AtomType::dispersion_imag)
+    .def("__repr__", [](const AtomType& self) {
+        return "<gemmi.SmallStructure.AtomType " + self.symbol + ">";
+    });
+
+  small_structure
+    .def(py::init<>())
+    .def_readwrite("name", &SmallStructure::name)
+    .def_readwrite("cell", &SmallStructure::cell)
+    .def_readwrite("spacegroup_hm", &SmallStructure::spacegroup_hm)
+    .def_readonly("sites", &SmallStructure::sites)
+    .def_readonly("atom_types", &SmallStructure::atom_types)
+    .def_readwrite("wavelength", &SmallStructure::wavelength)
+    .def("add_site", [](SmallStructure& self, const SmallStructure::Site& site) {
+        self.sites.push_back(site);
+    })
+    .def("find_spacegroup", &SmallStructure::find_spacegroup)
+    .def("get_atom_type", &SmallStructure::get_atom_type)
+    .def("get_all_unit_cell_sites", &SmallStructure::get_all_unit_cell_sites)
+    .def("remove_hydrogens", &SmallStructure::remove_hydrogens)
+    .def("change_occupancies_to_crystallographic",
+         &SmallStructure::change_occupancies_to_crystallographic,
+         py::arg("max_dist")=0.4)
+    .def("setup_cell_images", &SmallStructure::setup_cell_images)
+    .def("make_cif_block", &make_cif_block_from_small_structure)
+    .def("__repr__", [](const SmallStructure& self) {
+        return "<gemmi.SmallStructure: " + std::string(self.name) + ">";
+    });
+  m.def("mx_to_sx_structure", &gemmi::mx_to_sx_structure,
+        py::arg("st"), py::arg("n")=0);
+}
+
 // used in cif.cpp
 void cif_parse_string(cif::Document& doc, const std::string& data) {
   tao::pegtl::memory_input<> in(data, "string");
