@@ -2,10 +2,12 @@
 
 #include "gemmi/topo.hpp"
 #include "gemmi/placeh.hpp"  // for adjust_hydrogen_distances
+#include "gemmi/crd.hpp"     // for prepare_crd, prepare_rst
 
 #include "common.h"
 #include <pybind11/stl.h>
 #include <pybind11/stl_bind.h>
+#include <pybind11/iostream.h>  // for detail::pythonbuf
 
 namespace py = pybind11;
 using namespace gemmi;
@@ -15,17 +17,13 @@ PYBIND11_MAKE_OPAQUE(std::vector<Topo::Angle>)
 PYBIND11_MAKE_OPAQUE(std::vector<Topo::Torsion>)
 PYBIND11_MAKE_OPAQUE(std::vector<Topo::Chirality>)
 PYBIND11_MAKE_OPAQUE(std::vector<Topo::Plane>)
-PYBIND11_MAKE_OPAQUE(std::vector<Topo::ExtraLink>)
-PYBIND11_MAKE_OPAQUE(std::vector<Topo::ChainInfo>)
-PYBIND11_MAKE_OPAQUE(std::vector<Topo::ResInfo::Prev>)
-PYBIND11_MAKE_OPAQUE(std::vector<Topo::ResInfo>)
-PYBIND11_MAKE_OPAQUE(std::vector<Topo::Rule>)
+PYBIND11_MAKE_OPAQUE(std::vector<Topo::Link>)
 
 void add_topo(py::module& m) {
   py::class_<Topo> topo(m, "Topo");
 
   py::enum_<HydrogenChange>(m, "HydrogenChange")
-    .value("None", HydrogenChange::None)
+    .value("NoChange", HydrogenChange::NoChange)
     .value("Shift", HydrogenChange::Shift)
     .value("Remove", HydrogenChange::Remove)
     .value("ReAdd", HydrogenChange::ReAdd)
@@ -62,63 +60,19 @@ void add_topo(py::module& m) {
     .def_readonly("atoms", &Topo::Plane::atoms)
     .def("has", &Topo::Plane::has)
     ;
-  py::enum_<Topo::Provenance>(m, "Provenance")
-    .value("None", Topo::Provenance::None)
-    .value("PrevLink", Topo::Provenance::PrevLink)
-    .value("Monomer", Topo::Provenance::Monomer)
-    .value("NextLink", Topo::Provenance::NextLink)
-    .value("ExtraLink", Topo::Provenance::ExtraLink)
+  py::class_<Topo::Link>(topo, "Link")
+    .def_readonly("link_id", &Topo::Link::link_id)
+    .def_readonly("res1", &Topo::Link::res1)
+    .def_readonly("res2", &Topo::Link::res2)
+    .def_readonly("alt1", &Topo::Link::alt1)
+    .def_readonly("alt2", &Topo::Link::alt2)
     ;
-  py::enum_<Topo::RKind>(m, "RKind")
-    .value("Bond", Topo::RKind::Bond)
-    .value("Angle", Topo::RKind::Angle)
-    .value("Torsion", Topo::RKind::Torsion)
-    .value("Chirality", Topo::RKind::Chirality)
-    .value("Plane", Topo::RKind::Plane)
-    ;
-  py::class_<Topo::Rule>(topo, "Rule")
-    .def_readonly("provenance", &Topo::Rule::provenance)
-    .def_readonly("rkind", &Topo::Rule::rkind)
-    .def_readonly("index", &Topo::Rule::index)
-    ;
-  py::class_<Topo::ResInfo> resinfo(topo, "ResInfo");
-  py::class_<Topo::ResInfo::Prev>(resinfo, "Prev")
-    .def_readonly("link", &Topo::ResInfo::Prev::link)
-    .def_readonly("idx", &Topo::ResInfo::Prev::idx)
-    .def("get", (Topo::ResInfo* (Topo::ResInfo::Prev::*)(Topo::ResInfo*)const) &Topo::ResInfo::Prev::get, py::return_value_policy::reference_internal)
-    ;
-  resinfo
-    .def_readonly("res", &Topo::ResInfo::res)
-    .def_readonly("prev", &Topo::ResInfo::prev)
-    .def_readonly("mods", &Topo::ResInfo::mods)
-    .def_readonly("chemcomp", &Topo::ResInfo::chemcomp)
-    .def_readonly("rules", &Topo::ResInfo::rules)
-    ;
-  py::class_<Topo::ChainInfo>(topo, "ChainInfo")
-    .def_readonly("name", &Topo::ChainInfo::name)
-    .def_readonly("entity_id", &Topo::ChainInfo::entity_id)
-    .def_readonly("polymer", &Topo::ChainInfo::polymer)
-    .def_readonly("polymer_type", &Topo::ChainInfo::polymer_type)
-    .def_readonly("res_infos", &Topo::ChainInfo::res_infos)
-    ;
-  py::class_<Topo::ExtraLink>(topo, "ExtraLink")
-    .def_readonly("res1", &Topo::ExtraLink::res1)
-    .def_readonly("res2", &Topo::ExtraLink::res2)
-    .def_readonly("alt1", &Topo::ExtraLink::alt1)
-    .def_readonly("alt2", &Topo::ExtraLink::alt2)
-    .def_readonly("link_id", &Topo::ExtraLink::link_id)
-    ;
-
   py::bind_vector<std::vector<Topo::Bond>>(m, "TopoBonds");
   py::bind_vector<std::vector<Topo::Angle>>(m, "TopoAngles");
   py::bind_vector<std::vector<Topo::Torsion>>(m, "TopoTorsions");
   py::bind_vector<std::vector<Topo::Chirality>>(m, "TopoChirs");
   py::bind_vector<std::vector<Topo::Plane>>(m, "TopoPlanes");
-  py::bind_vector<std::vector<Topo::ChainInfo>>(m, "TopoChainInfos");
-  py::bind_vector<std::vector<Topo::Rule>>(m, "TopoRules");
-  py::bind_vector<std::vector<Topo::ResInfo::Prev>>(m, "TopoResInfoPrevs");
-  py::bind_vector<std::vector<Topo::ResInfo>>(m, "TopoResInfos");
-  py::bind_vector<std::vector<Topo::ExtraLink>>(m, "TopoExtraLinks");
+  py::bind_vector<std::vector<Topo::Link>>(m, "TopoLinks");
 
   topo
     .def(py::init<>())
@@ -130,12 +84,37 @@ void add_topo(py::module& m) {
     .def_readonly("chirs", &Topo::chirs)
     .def_readonly("planes", &Topo::planes)
     .def_readonly("extras", &Topo::extras)
-    .def_readonly("chain_infos", &Topo::chain_infos)
     .def("ideal_chiral_abs_volume", &Topo::ideal_chiral_abs_volume)
+    .def("links_to_previous", [](Topo& self, Residue* res) {
+        if (Topo::ResInfo* ri = self.find_resinfo(res))
+          return ri->prev;
+        fail("links_to_previous(): Residue not found");
+    }, py::return_value_policy::reference_internal)
+    .def("first_bond_in_link", &Topo::first_bond_in_link,
+         py::return_value_policy::reference_internal)
     ;
 
-  m.def("prepare_topology", &prepare_topology,
-        py::arg("st"), py::arg("monlib"), py::arg("model_index")=0,
-        py::arg("h_change")=HydrogenChange::None, py::arg("reorder")=false,
-        py::arg("raise_errors")=false);
+  m.def("prepare_topology",
+    [](Structure& st, MonLib& monlib, size_t model_index,
+       HydrogenChange h_change, bool reorder,
+       const py::object& pywarnings, bool ignore_unknown_links) {
+      std::ostream* warnings = nullptr;
+      std::ostream os(nullptr);
+      std::unique_ptr<py::detail::pythonbuf> buffer;
+      if (!pywarnings.is_none()) {
+        buffer.reset(new py::detail::pythonbuf(pywarnings));
+        os.rdbuf(buffer.get());
+        warnings = &os;
+      }
+      return prepare_topology(st, monlib, model_index, h_change, reorder,
+                              warnings, ignore_unknown_links);
+    }, py::arg("st"), py::arg("monlib"), py::arg("model_index")=0,
+       py::arg("h_change")=HydrogenChange::NoChange, py::arg("reorder")=false,
+       py::arg("warnings")=py::none(), py::arg("ignore_unknown_links")=false);
+
+  // crd.hpp
+  m.def("prepare_crd", &prepare_crd);
+  m.def("prepare_rst", &prepare_rst);
+  m.def("prepare_refmac_crd", &prepare_refmac_crd);
+  m.def("add_automatic_links", &add_automatic_links);
 }
