@@ -1,8 +1,9 @@
 // Copyright 2017 Global Phasing Ltd.
 
 #include "gemmi/unitcell.hpp"
-#include "gemmi/tostr.hpp"  // tostr
-#include "gemmi/gruber.hpp"  // GruberVector
+#include "tostr.hpp"  // tostr
+#include "gemmi/cellred.hpp"  // GruberVector
+#include "gemmi/twin.hpp"  // find_lattice_2fold_ops
 
 #include <cstdio>  // for snprintf
 #include <array>
@@ -57,6 +58,7 @@ template<typename T> void add_smat33(py::module& m, const char* name) {
     .def("nonzero", &M::nonzero)
     .def("determinant", &M::determinant)
     .def("inverse", &M::inverse)
+    .def("added_kI", &M::added_kI)
     .def("r_u_r", (double (M::*)(const Vec3&) const) &M::r_u_r)
     .def("r_u_r", [](const M& self, py::array_t<int> arr) {
         int nrow = (int) arr.shape(0);
@@ -70,7 +72,10 @@ template<typename T> void add_smat33(py::module& m, const char* name) {
            v.push_back((T)self.r_u_r(Vec3(r(row, 0), r(row, 1), r(row, 2))));
         return py_array_from_vector(std::move(v));
     }, py::arg().noconvert())
-    .def("transformed_by", &M::template transformed_by<double>)
+    .def("multiply", &M::multiply)
+    .def(py::self + py::self)
+    .def(py::self - py::self)
+    .def("transformed_by", &M::template transformed_by<T>)
     .def("calculate_eigenvalues", &M::calculate_eigenvalues)
     .def("__repr__", [name](const M& m) {
         return tostr("<gemmi.", name, '(', m.u11, ", ", m.u22, ", ", m.u33, ", ",
@@ -141,10 +146,14 @@ void add_unitcell(py::module& m) {
     })
     .def("row_copy", &Mat33::row_copy)
     .def("column_copy", &Mat33::column_copy)
+    .def(py::self + py::self)
+    .def(py::self - py::self)
     .def("multiply", (Mat33 (Mat33::*)(const Mat33&) const) &Mat33::multiply)
     .def("multiply", (Vec3 (Mat33::*)(const Vec3&) const) &Mat33::multiply)
     .def("left_multiply", &Mat33::left_multiply)
+    .def("multiply_by_diagonal", &Mat33::multiply_by_diagonal)
     .def("transpose", &Mat33::transpose)
+    .def("trace", &Mat33::trace)
     .def("approx", &Mat33::approx, py::arg("other"), py::arg("epsilon"))
     .def("determinant", &Mat33::determinant)
     .def("inverse", &Mat33::inverse)
@@ -166,7 +175,8 @@ void add_unitcell(py::module& m) {
   add_smat33<double>(m, "SMat33d");
   add_smat33<float>(m, "SMat33f");
 
-  py::class_<Transform>(m, "Transform")
+  py::class_<Transform> transform(m, "Transform");
+  transform
     .def(py::init<>())
     .def(py::init([](const Mat33& m, const Vec3& v) {
       Transform* tr = new Transform();
@@ -178,8 +188,10 @@ void add_unitcell(py::module& m) {
     .def_readonly("vec", &Transform::vec)
     .def("inverse", &Transform::inverse)
     .def("apply", &Transform::apply)
+    .def("combine", &Transform::combine)
     .def("is_identity", &Transform::is_identity)
-    .def("approx", &Transform::approx);
+    .def("approx", &Transform::approx, py::arg("other"), py::arg("epsilon"));
+  transform.attr("__matmul__") = transform.attr("combine");
 
   py::class_<Correlation>(m, "Correlation")
     .def_readonly("n", &Correlation::n)
@@ -223,12 +235,19 @@ void add_unitcell(py::module& m) {
     .def("apply", &FTransform::apply);
 
 
-  py::class_<SymImage>(m, "SymImage")
-    .def("dist", &SymImage::dist)
-    .def("__repr__", [](const SymImage& self) {
-        return "<gemmi.SymImage box:[" +
-          triple(self.box[0], self.box[1], self.box[2]) +
-          "] sym:" + std::to_string(self.sym_id) + ">";
+  py::class_<NearestImage>(m, "NearestImage")
+    .def("dist", &NearestImage::dist)
+    .def("symmetry_code", &NearestImage::symmetry_code, py::arg("underscore")=true)
+    .def_readonly("sym_idx", &NearestImage::sym_idx)
+    .def_property_readonly("pbc_shift", [](const NearestImage& self) {
+        return py::make_tuple(self.pbc_shift[0], self.pbc_shift[1], self.pbc_shift[2]);
+    })
+    .def("__repr__", [](const NearestImage& self) {
+        using namespace std;  // VS2015/17 doesn't like std::snprintf
+        char buf[64];
+        snprintf(buf, 64, "<gemmi.NearestImage %s in distance %.2f>",
+                 self.symmetry_code(true).c_str(), self.dist());
+        return std::string(buf);
     });
 
   py::enum_<Asu>(m, "Asu")
@@ -250,6 +269,9 @@ void add_unitcell(py::module& m) {
     .def_readonly("volume", &UnitCell::volume)
     .def_readonly("images", &UnitCell::images)
     .def_property_readonly("parameters", &make_parameters_tuple)
+    .def_readonly("frac", &UnitCell::frac)
+    .def_readonly("orth", &UnitCell::orth)
+    // These two functions are deprecated, use frac.mat and orth.mat.
     .def_property_readonly("fractionalization_matrix",
                            [](const UnitCell& self) { return self.frac.mat; })
     .def_property_readonly("orthogonalization_matrix",
@@ -262,13 +284,24 @@ void add_unitcell(py::module& m) {
     .def("is_compatible_with_spacegroup", &UnitCell::is_compatible_with_spacegroup,
          py::arg("sg"), py::arg("eps")=1e-3)
     .def("is_crystal", &UnitCell::is_crystal)
+    .def("approx", &UnitCell::approx, py::arg("other"), py::arg("epsilon"))
+    .def("is_similar", &UnitCell::is_similar,
+         py::arg("other"), py::arg("rel"), py::arg("deg"))
     .def("calculate_u_eq", &UnitCell::calculate_u_eq)
     .def("fractionalize", &UnitCell::fractionalize)
     .def("orthogonalize", &UnitCell::orthogonalize)
+    .def("orthogonalize_box", &UnitCell::orthogonalize_box)
+    .def("op_as_transform", &UnitCell::op_as_transform)
     .def("volume_per_image", &UnitCell::volume_per_image)
     .def("find_nearest_image", &UnitCell::find_nearest_image,
          py::arg("ref"), py::arg("pos"), py::arg("asu")=Asu::Any)
-    .def("find_nearest_pbc_image", &UnitCell::find_nearest_pbc_image,
+    .def("find_nearest_pbc_image",
+         (NearestImage (UnitCell::*)(const Fractional&, Fractional, int) const)
+         &UnitCell::find_nearest_pbc_image,
+         py::arg("fref"), py::arg("fpos"), py::arg("image_idx"))
+    .def("find_nearest_pbc_image",
+         (NearestImage (UnitCell::*)(const Position&, const Position&, int) const)
+         &UnitCell::find_nearest_pbc_image,
          py::arg("ref"), py::arg("pos"), py::arg("image_idx"))
     .def("find_nearest_pbc_position", &UnitCell::find_nearest_pbc_position,
          py::arg("ref"), py::arg("pos"), py::arg("image_idx"), py::arg("inverse")=false)
@@ -292,7 +325,7 @@ void add_unitcell(py::module& m) {
     .def("reciprocal_metric_tensor", &UnitCell::reciprocal_metric_tensor)
     .def("reciprocal", &UnitCell::reciprocal)
     .def("get_hkl_limits", &UnitCell::get_hkl_limits, py::arg("dmin"))
-    .def("g6", &UnitCell::g6, py::arg("centring_type"))
+    .def("primitive_orth_matrix", &UnitCell::primitive_orth_matrix, py::arg("centring_type"))
     .def(py::self == py::self)
     .def("__repr__", [](const UnitCell& self) {
         return "<gemmi.UnitCell(" + triple(self.a, self.b, self.c)
@@ -306,12 +339,14 @@ void add_unitcell(py::module& m) {
          }
     ));
 
+  py::class_<SellingVector> selling_vector(m, "SellingVector");
+
   py::class_<GruberVector>(m, "GruberVector")
     .def(py::init<const std::array<double,6>&>())
-    .def(py::init([](const UnitCell& u, const SpaceGroup* sg) {
-        return new GruberVector(u.g6(sg ? sg->centring_type() : 'P'));
-    }))
-    .def("parameters", &GruberVector::parameters)
+    .def(py::init<const UnitCell&, const SpaceGroup*, bool>(),
+         py::arg("cell"), py::arg("sg"), py::arg("track_change_of_basis")=false)
+    .def(py::init<const UnitCell&, char, bool>(),
+         py::arg("cell"), py::arg("centring"), py::arg("track_change_of_basis")=false)
     .def_property_readonly("parameters", [](const GruberVector& g) {
       return make_six_tuple(g.parameters());
     })
@@ -319,7 +354,11 @@ void add_unitcell(py::module& m) {
       return make_six_tuple(self.cell_parameters());
     })
     .def("get_cell",
-         [](const GruberVector& self) { return new UnitCell(self.cell_parameters()); })
+         [](const GruberVector& self) { return new UnitCell(self.get_cell()); })
+    .def_property_readonly("change_of_basis", [](const GruberVector& self) {
+        return self.change_of_basis.get();
+    }, py::return_value_policy::reference_internal)
+    .def("selling", &GruberVector::selling)
     .def("is_normalized", &GruberVector::is_normalized)
     .def("is_buerger", &GruberVector::is_buerger, py::arg("epsilon")=1e-9)
     .def("normalize", &GruberVector::normalize, py::arg("epsilon")=1e-9)
@@ -335,4 +374,38 @@ void add_unitcell(py::module& m) {
                  self.A, self.B, self.C, self.xi, self.eta, self.zeta);
         return std::string(buf);
     });
+
+  selling_vector
+    .def(py::init<const std::array<double,6>&>())
+    .def(py::init<const UnitCell&, const SpaceGroup*>())
+    .def_property_readonly("parameters", [](const SellingVector& self) {
+      return make_six_tuple(self.s);
+    })
+    .def("cell_parameters", [](const SellingVector& self) {
+      return make_six_tuple(self.cell_parameters());
+    })
+    .def("get_cell",
+         [](const SellingVector& self) { return new UnitCell(self.get_cell()); })
+    .def("sum_b_squared", &SellingVector::sum_b_squared)
+    .def("gruber", &SellingVector::gruber)
+    .def("is_reduced", &SellingVector::is_reduced, py::arg("epsilon")=1e-9)
+    .def("reduce_step", &SellingVector::reduce_step, py::arg("epsilon")=1e-9)
+    .def("reduce", &SellingVector::reduce,
+         py::arg("epsilon")=1e-9, py::arg("iteration_limit")=100)
+    .def("sort", &SellingVector::sort, py::arg("epsilon")=1e-9)
+    .def("__repr__", [](const SellingVector& self) {
+        using namespace std;  // VS2015/17 doesn't like std::snprintf
+        char buf[256];
+        snprintf(buf, 256, "<gemmi.SellingVector((%.2f, %.2f, %.2f, %.2f, %.2f, %.2f))>",
+                 self.s[0], self.s[1], self.s[2], self.s[3], self.s[4], self.s[5]);
+        return std::string(buf);
+    });
+
+  m.def("find_lattice_2fold_ops", &find_lattice_2fold_ops,
+        py::arg("reduced_cell"), py::arg("max_obliq"));
+  m.def("find_lattice_symmetry_r", &find_lattice_symmetry_r);
+  m.def("find_lattice_symmetry", &find_lattice_symmetry,
+        py::arg("cell"), py::arg("centring"), py::arg("max_obliq"));
+  m.def("find_twin_laws", &find_twin_laws,
+        py::arg("cell"), py::arg("sg"), py::arg("max_obliq"), py::arg("all_ops"));
 }
